function VarTypeReference(
    table,      // Responsible TableOfVars instance
    varName,    // Name of this variable
    varType,    // Name of (base) type
    dimensions  // Optional: Number of dimensions (default 0, greater 0 means array)
) {
    this.baseType = (varName == 'this') ? true : varType;
    if (dimensions)
        this.dimensions = dimensions;
    VarType.call(this, table, varName);
}
VarTypeReference.prototype = Object.create(VarType.prototype);
VarTypeReference.prototype.constructor = VarTypeReference;

/** Used number of bytes to store this variable */
VarTypeReference.countBytes = 4;

/**
* Add table rows to this.table
* @return undefined
*/
VarTypeReference.prototype.createTr = function() {
    const isThis = (this.varName == 'this');
    this._tr = this.table.addTrVar(
        isThis ? ScoTools.createHtml('span', this.table.getDocument().createElement('code'), { 'class': 'hljs-keyword' }, 'this').parentNode : this.varName,
        isThis ? null : this.table.createCellContentType(this.table.getDocument(), this.baseType, this.dimensions),
        this.input,
        isThis ? undefined : (() => this.remove())
    );
}

/** Name of variable's base type */
VarTypeReference.prototype.baseType = '';

/** Number of dimensions (default 0, greater 0 means array) */
VarTypeReference.prototype.dimensions = 0;

/** Name of variable type */
VarTypeReference.typeName = null;
Object.defineProperty(VarTypeReference.prototype, 'typeName', {
    enumerable: true,
    get() { return this.baseType + '[]'.repeat(this.dimensions); }
});

/** Default value of variables of this type */
VarTypeReference.defaultValue = null;

/** Present value */
Object.defineProperty(VarTypeReference.prototype, 'value', { enumerable: true,
    get() { return this.input.options[Math.max(0, this.input.selectedIndex)].textContent; },
    set(v) {
        const options = Array.from(this.input.options).map(node => node.textContent);
        const index = (v == 'null') ? 0 : options.indexOf(v);
        this.input.selectedIndex = ((index < 1) || (options[index].substring(0, 4) == 'new ')) ? 0 : index;
    }
});

/**
* Create a input for this type
* @return HTMLElement       Created input, select, …
*/
VarTypeReference.prototype._createInput = function(
    document        // HTMLDocument instance to use
) {
    const input = document.createElement('select');
    ScoTools.createHtml('option', input, null, 'null');
    const postfix = (this.dimensions >= 1) ? ('[]'.repeat(this.dimensions - 1) + '[…]') : '()';
    const allClasses = this.table.classes;
    const classInfo = allClasses.filter(({ className }) => className == this.baseType).reduce((a, b) => b || a, {});
    const isa = classInfo.isa ? classInfo.isa : [];
    isa.push(this.baseType);
    allClasses
        .filter(({ className, abstract }) => !abstract && (isa.indexOf(className) > -1))
        .forEach(({ className }) => ScoTools.createHtml('option', input, null, 'new ' + className + postfix));
    this.table.getInstancesOf(this.baseType, this.dimensions).forEach(address => ScoTools.createHtml('option', input, null, address));
    input.selectedIndex = 0;
    input.addEventListener('change', () => {
        const selected = input.options[input.selectedIndex].textContent;
        if (selected.substring(0, 4) != 'new ')
            return;
        const className = selected.substring(4, selected.length - postfix.length);
        const created = this.table.createInstancesOf(className, this.dimensions);
        const index = Array.from(input.options).slice(2).map(option => option.textContent).indexOf(created) + 2;
        input.selectedIndex = (index > 1) ? index : 0;
    });
    return input;
}

/**
* Add another instance to list of possible values
* @return undefined
*/
VarTypeReference.prototype.addInstance = function(
    address     // Human readable address of instance to add
) {
    if (Array.from(this.input.options).slice(2).filter(({ text }) => (text == address)).length)
        return;
    const value = this.value;
    const options = Array.from(this.input.options).map(node => node.textContent);
    let countNoInstances = 1;
    while ((countNoInstances < options.length) && (options[countNoInstances].substring(0, 4) == 'new '))
        ++countNoInstances;
    const insertAtIndex = Array.from(this.input.options).slice(countNoInstances).findIndex(option => (address < option.textContent)) + countNoInstances;
    if (insertAtIndex == countNoInstances - 1)
        ScoTools.createHtml('option', this.input, null, address);
    else
        ScoTools.createHtml('option', this.input, null, address, this.input.childNodes[insertAtIndex]);
    this.value = value;
}

/**
* Remove an instance to list of possible values
* @return undefined
*/
VarTypeReference.prototype.removeInstance = function(
    address     // Human readable address of instance to remove
) {
    const value = this.value;
    Array.from(this.input.options).filter(option => (option.textContent == address)).forEach(option => option.parentNode.removeChild(option));
    this.input.selectedIndex = Math.max(0, Array.from(this.input.options).map(option => option.textContent).indexOf(value));
}

/**
* Convert contained data to JSON
* @return Object        JSON representation of contained data
*/
VarTypeReference.prototype.toJSON = function() {
    if (this.varName == 'this')
        return ({ key: 'this', ref: this.table.addressToPointer(this.value) });
    const json = {
        key: this.varName,
        type: this.baseType,
        ref: this.table.addressToPointer(this.value)
    };
    if (this.dimensions)
        json.dimensions = this.dimensions;
    return json;
}
